@@ -48,7 +48,7 @@ import dataclasses
 import itertools
 import functools
 import subprocess
-from typing import Any, List, Dict, Optional, Iterator, TYPE_CHECKING
+from typing import Any, List, Dict, Optional, Iterator, Type, TYPE_CHECKING
 
 from PyQt5.QtCore import (Qt, QObject, QVariant, QMetaType, QByteArray, pyqtSlot,
                           pyqtSignal, QTimer, QProcess, QUrl)
@@ -65,7 +65,7 @@ if TYPE_CHECKING:
 
 from qutebrowser.config import config
 from qutebrowser.misc import objects
-from qutebrowser.utils import qtutils, log, utils, debug, message, version
+from qutebrowser.utils import qtutils, log, utils, debug, message, version, objreg
 from qutebrowser.qt import sip
 
 
@@ -103,6 +103,45 @@ class Error(Exception):
     """Raised when something goes wrong with notifications."""
 
 
+class DBusError(Error):
+    """Raised when there was an error coming from DBus."""
+
+    _NON_FATAL_ERRORS = {
+        # notification daemon is gone
+        "org.freedesktop.DBus.Error.NoReply",
+
+        # https://gitlab.gnome.org/GNOME/gnome-flashback/-/blob/3.40.0/gnome-flashback/libnotifications/nd-daemon.c#L178-187
+        # Exceeded maximum number of notifications
+        "org.freedesktop.Notifications.MaxNotificationsExceeded",
+
+        # https://bugs.kde.org/show_bug.cgi?id=409157
+        # https://github.com/KDE/plasma-workspace/blob/v5.21.4/libnotificationmanager/server_p.cpp#L227-L237
+        # Created too many similar notifications in quick succession
+        "org.freedesktop.Notifications.Error.ExcessNotificationGeneration",
+
+        # From https://crashes.qutebrowser.org/view/b8c9838a
+        # Process org.freedesktop.Notifications received signal 5
+        # probably when notification daemon crashes?
+        "org.freedesktop.DBus.Error.Spawn.ChildSignaled",
+
+        # https://crashes.qutebrowser.org/view/f76f58ae
+        # Process org.freedesktop.Notifications exited with status 1
+        "org.freedesktop.DBus.Error.Spawn.ChildExited",
+
+        # https://crashes.qutebrowser.org/view/8889d0b5
+        # Could not activate remote peer.
+        "org.freedesktop.DBus.Error.NameHasNoOwner",
+    }
+
+    def __init__(self, msg: QDBusMessage) -> None:
+        assert msg.type() == QDBusMessage.ErrorMessage
+        self.error = msg.errorName()
+        self.error_message = msg.errorMessage()
+        self.is_fatal = self.error not in self._NON_FATAL_ERRORS
+        text = f"{self.error}: {self.error_message}"
+        super().__init__(text)
+
+
 class AbstractNotificationAdapter(QObject):
 
     """An adapter taking notifications and displaying them.
@@ -188,31 +227,7 @@ class NotificationBridgePresenter(QObject):
             message.error("Can't switch to qt notification presenter at runtime.")
             setting = "auto"
 
-        if setting in ["auto", "libnotify"]:
-            candidates = [
-                DBusNotificationAdapter,
-                SystrayNotificationAdapter,
-                MessagesNotificationAdapter,
-            ]
-        elif setting == "systray":
-            candidates = [
-                SystrayNotificationAdapter,
-                DBusNotificationAdapter,
-                MessagesNotificationAdapter,
-            ]
-        elif setting == "herbe":
-            candidates = [
-                HerbeNotificationAdapter,
-                DBusNotificationAdapter,
-                SystrayNotificationAdapter,
-                MessagesNotificationAdapter,
-            ]
-        elif setting == "messages":
-            candidates = [MessagesNotificationAdapter]  # always succeeds
-        else:
-            raise utils.Unreachable(setting)
-
-        for candidate in candidates:
+        for candidate in self._get_adapter_candidates(setting):
             try:
                 self._adapter = candidate()
             except Error as e:
@@ -231,6 +246,32 @@ class NotificationBridgePresenter(QObject):
         self._adapter.error.connect(self._on_adapter_error)
         self._adapter.clear_all.connect(self._on_adapter_clear_all)
 
+    def _get_adapter_candidates(
+        self,
+        setting: str,
+    ) -> List[Type[AbstractNotificationAdapter]]:
+        candidates: Dict[str, List[Type[AbstractNotificationAdapter]]] = {
+            "libnotify": [
+                DBusNotificationAdapter,
+                SystrayNotificationAdapter,
+                MessagesNotificationAdapter,
+            ],
+            "systray": [
+                SystrayNotificationAdapter,
+                DBusNotificationAdapter,
+                MessagesNotificationAdapter,
+            ],
+            "herbe": [
+                HerbeNotificationAdapter,
+                DBusNotificationAdapter,
+                SystrayNotificationAdapter,
+                MessagesNotificationAdapter,
+            ],
+            "messages": [MessagesNotificationAdapter],  # always succeeds
+        }
+        candidates["auto"] = candidates["libnotify"]
+        return candidates[setting]
+
     def install(self, profile: "QWebEngineProfile") -> None:
         """Set the profile to use this bridge as the presenter."""
         # WORKAROUND for
@@ -276,9 +317,6 @@ class NotificationBridgePresenter(QObject):
             log.misc.debug("Adapter vanished, bailing out")  # type: ignore[unreachable]
             return
 
-        if notification_id <= 0:
-            raise Error(f"Got invalid notification id {notification_id}")
-
         if replaces_id is None:
             if notification_id in self._active_notifications:
                 raise Error(f"Got duplicate id {notification_id}")
@@ -368,6 +406,17 @@ class NotificationBridgePresenter(QObject):
             # https://www.riverbankcomputing.com/pipermail/pyqt/2020-May/042918.html
             log.misc.debug(f"Ignoring click request for notification {notification_id} "
                            "due to PyQt bug")
+            return
+        self._focus_first_matching_tab(notification)
+
+    def _focus_first_matching_tab(self, notification: "QWebEngineNotification") -> None:
+        for win_id in objreg.window_registry:
+            tabbedbrowser = objreg.get("tabbed-browser", window=win_id, scope="window")
+            for idx, tab in enumerate(tabbedbrowser.widgets()):
+                if tab.url().matches(notification.origin(), QUrl.RemovePath):
+                    tabbedbrowser.widget.setCurrentIndex(idx)
+                    return
+        log.misc.debug(f"No matching tab found for {notification.origin()}")
 
     def _drop_adapter(self) -> None:
         """Drop the currently active adapter (if any).
@@ -611,17 +660,18 @@ class HerbeNotificationAdapter(AbstractNotificationAdapter):
         so there's no point.
         """
         if status == QProcess.CrashExit:
-            return
-
-        if code == 0:
+            pass
+        elif code == 0:
             self.click_id.emit(pid)
         elif code == 2:
-            self.close_id.emit(pid)
+            pass
         else:
             proc = self.sender()
             stderr = proc.readAllStandardError()
             raise Error(f'herbe exited with status {code}: {stderr}')
 
+        self.close_id.emit(pid)
+
     @pyqtSlot(QProcess.ProcessError)
     def _on_error(self, error: QProcess.ProcessError) -> None:
         if error == QProcess.Crashed:
@@ -654,6 +704,7 @@ class _ServerQuirks:
     skip_capabilities: bool = False
     wrong_replaces_id: bool = False
     no_padded_images: bool = False
+    wrong_closes_type: bool = False
 
 
 @dataclasses.dataclass
@@ -703,22 +754,8 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
     SPEC_VERSION = "1.2"  # Released in January 2011, still current in March 2021.
     NAME = "libnotify"
 
-    _NON_FATAL_ERRORS = {
-        # notification daemon is gone
-        "org.freedesktop.DBus.Error.NoReply",
-
-        # https://gitlab.gnome.org/GNOME/gnome-flashback/-/blob/3.40.0/gnome-flashback/libnotifications/nd-daemon.c#L178-187
-        # Exceeded maximum number of notifications
-        "org.freedesktop.Notifications.MaxNotificationsExceeded",
-
-        # https://bugs.kde.org/show_bug.cgi?id=409157
-        # https://github.com/KDE/plasma-workspace/blob/v5.21.4/libnotificationmanager/server_p.cpp#L227-L237
-        # Created too many similar notifications in quick succession
-        "org.freedesktop.Notifications.Error.ExcessNotificationGeneration",
-    }
-
     def __init__(self, parent: QObject = None) -> None:
-        super().__init__(bridge)
+        super().__init__(parent)
         assert _notifications_supported()
 
         if utils.is_windows:
@@ -744,7 +781,7 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
             self._on_service_unregistered)
 
         test_service = 'test-notification-service' in objects.debug_flags
-        service = self.TEST_SERVICE if test_service else self.SERVICE
+        service = f"{self.TEST_SERVICE}{os.getpid()}" if test_service else self.SERVICE
 
         self.interface = QDBusInterface(service, self.PATH, self.INTERFACE, bus)
         if not self.interface.isValid():
@@ -815,8 +852,9 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
             # https://github.com/sboli/twmn/pull/96
             return _ServerQuirks(spec_version="0")
         elif (name, vendor) == ("tiramisu", "Sweets"):
-            # https://github.com/Sweets/tiramisu/issues/20
-            return _ServerQuirks(skip_capabilities=True)
+            if utils.VersionNumber.parse(ver) < utils.VersionNumber(2, 0):
+                # https://github.com/Sweets/tiramisu/issues/20
+                return _ServerQuirks(skip_capabilities=True)
         elif (name, vendor) == ("lxqt-notificationd", "lxqt.org"):
             quirks = _ServerQuirks()
             parsed_version = utils.VersionNumber.parse(ver)
@@ -840,12 +878,21 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
                 wrong_replaces_id=True,
             )
         elif (name, vendor) == ("Raven", "Budgie Desktop Developers"):
+            # Before refactor
             return _ServerQuirks(
                 # https://github.com/solus-project/budgie-desktop/issues/2114
                 escape_title=True,
                 # https://github.com/solus-project/budgie-desktop/issues/2115
                 wrong_replaces_id=True,
             )
+        elif (name, vendor) == (
+                "Budgie Notification Server", "Budgie Desktop Developers"):
+            # After refactor: https://github.com/BuddiesOfBudgie/budgie-desktop/pull/36
+            if utils.VersionNumber.parse(ver) < utils.VersionNumber(10, 6, 2):
+                return _ServerQuirks(
+                    # https://github.com/BuddiesOfBudgie/budgie-desktop/issues/118
+                    wrong_closes_type=True,
+                )
         return None
 
     def _get_server_info(self) -> None:
@@ -900,12 +947,7 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
         ], expected_type
 
         if msg.type() == QDBusMessage.ErrorMessage:
-            err = msg.errorName()
-            if err in self._NON_FATAL_ERRORS:
-                self.error.emit(msg.errorMessage())
-                return
-
-            raise Error(f"Got DBus error: {err} - {msg.errorMessage()}")
+            raise DBusError(msg)
 
         signature = msg.signature()
         if signature != expected_signature:
@@ -921,7 +963,7 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
                 f"Got a message of type {type_str} but expected {expected_type_str}"
                 f"(args: {msg.arguments()})")
 
-    def present(
+    def present(  # noqa: C901 ("too complex")
         self,
         qt_notification: "QWebEngineNotification",
         *,
@@ -975,7 +1017,15 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
             hints,
             -1,  # timeout; -1 means 'use default'
         )
-        self._verify_message(reply, "u", QDBusMessage.ReplyMessage)
+
+        try:
+            self._verify_message(reply, "u", QDBusMessage.ReplyMessage)
+        except DBusError as e:
+            if e.is_fatal:
+                raise
+            self.error.emit(e.error_message)
+            # Return value gets ignored in NotificationBridgePresenter.present
+            return -1
 
         notification_id = reply.arguments()[0]
 
@@ -989,6 +1039,9 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
             else:
                 log.misc.error(msg)
 
+        if notification_id <= 0:
+            self.error.emit(f"Got invalid notification id {notification_id}")
+
         return notification_id
 
     def _convert_image(self, qimage: QImage) -> Optional[QDBusArgument]:
@@ -1067,7 +1120,13 @@ class DBusNotificationAdapter(AbstractNotificationAdapter):
     @pyqtSlot(QDBusMessage)
     def _handle_close(self, msg: QDBusMessage) -> None:
         """Handle NotificationClosed from DBus."""
-        self._verify_message(msg, "uu", QDBusMessage.SignalMessage)
+        try:
+            self._verify_message(msg, "uu", QDBusMessage.SignalMessage)
+        except Error:
+            if not self._quirks.wrong_closes_type:
+                raise
+            self._verify_message(msg, "ui", QDBusMessage.SignalMessage)
+
         notification_id, _close_reason = msg.arguments()
         self.close_id.emit(notification_id)
 
