@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2016-2019 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2016-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,7 +15,7 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Showing prompts above the statusbar."""
 
@@ -23,8 +23,9 @@ import os.path
 import html
 import collections
 import functools
+import dataclasses
+from typing import Deque, MutableSequence, Optional, cast
 
-import attr
 from PyQt5.QtCore import (pyqtSlot, pyqtSignal, Qt, QTimer, QDir, QModelIndex,
                           QItemSelectionModel, QObject, QEventLoop)
 from PyQt5.QtWidgets import (QWidget, QGridLayout, QVBoxLayout, QLineEdit,
@@ -32,23 +33,23 @@ from PyQt5.QtWidgets import (QWidget, QGridLayout, QVBoxLayout, QLineEdit,
                              QSpacerItem)
 
 from qutebrowser.browser import downloads
-from qutebrowser.config import config, configtypes, configexc
+from qutebrowser.config import config, configtypes, configexc, stylesheet
 from qutebrowser.utils import usertypes, log, utils, qtutils, objreg, message
 from qutebrowser.keyinput import modeman
 from qutebrowser.api import cmdutils
 from qutebrowser.utils import urlmatch
 
 
-prompt_queue = None
+prompt_queue = cast('PromptQueue', None)
 
 
-@attr.s
+@dataclasses.dataclass
 class AuthInfo:
 
     """Authentication info returned by a prompt."""
 
-    user = attr.ib()
-    password = attr.ib()
+    user: str
+    password: str
 
 
 class Error(Exception):
@@ -101,8 +102,8 @@ class PromptQueue(QObject):
         super().__init__(parent)
         self._question = None
         self._shutting_down = False
-        self._loops = []
-        self._queue = collections.deque()
+        self._loops: MutableSequence[qtutils.EventLoop] = []
+        self._queue: Deque[usertypes.Question] = collections.deque()
         message.global_bridge.mode_left.connect(self._on_mode_left)
 
     def __repr__(self):
@@ -193,9 +194,11 @@ class PromptQueue(QObject):
             loop.destroyed.connect(lambda: self._loops.remove(loop))
             question.completed.connect(loop.quit)
             question.completed.connect(loop.deleteLater)
-            log.prompt.debug("Starting loop.exec_() for {}".format(question))
-            loop.exec_(QEventLoop.ExcludeSocketNotifiers)
-            log.prompt.debug("Ending loop.exec_() for {}".format(question))
+            log.prompt.debug("Starting loop.exec() for {}".format(question))
+            flags = cast(QEventLoop.ProcessEventsFlags,
+                         QEventLoop.ExcludeSocketNotifiers)
+            loop.exec(flags)
+            log.prompt.debug("Ending loop.exec() for {}".format(question))
 
             log.prompt.debug("Restoring old question {}".format(old_question))
             self._question = old_question
@@ -265,6 +268,7 @@ class PromptContainer(QWidget):
         }
 
         QTreeView {
+            selection-color: {{ conf.colors.prompts.selected.fg }};
             selection-background-color: {{ conf.colors.prompts.selected.bg }};
             border: {{ conf.colors.prompts.border }};
         }
@@ -275,6 +279,7 @@ class PromptContainer(QWidget):
 
         QTreeView::item:selected, QTreeView::item:selected:hover,
         QTreeView::branch:selected {
+            color: {{ conf.colors.prompts.selected.fg }};
             background-color: {{ conf.colors.prompts.selected.bg }};
         }
     """
@@ -285,11 +290,11 @@ class PromptContainer(QWidget):
         self._layout = QVBoxLayout(self)
         self._layout.setContentsMargins(10, 10, 10, 10)
         self._win_id = win_id
-        self._prompt = None
+        self._prompt: Optional[_BasePrompt] = None
 
         self.setObjectName('PromptContainer')
         self.setAttribute(Qt.WA_StyledBackground, True)
-        config.set_register_stylesheet(self)
+        stylesheet.set_register(self)
 
         message.global_bridge.prompt_done.connect(self._on_prompt_done)
         prompt_queue.show_prompts.connect(self._on_show_prompts)
@@ -391,6 +396,7 @@ class PromptContainer(QWidget):
                    For boolean prompts, "yes"/"no" are accepted as value.
             save: Save the value to the config.
         """
+        assert self._prompt is not None
         question = self._prompt.question
 
         try:
@@ -418,6 +424,7 @@ class PromptContainer(QWidget):
                      cmdline.
             pdfjs: Open the download via PDF.js.
         """
+        assert self._prompt is not None
         try:
             self._prompt.download_open(cmdline, pdfjs=pdfjs)
         except UnsupportedOperationError:
@@ -432,6 +439,7 @@ class PromptContainer(QWidget):
         Args:
             which: 'next', 'prev'
         """
+        assert self._prompt is not None
         try:
             self._prompt.item_focus(which)
         except UnsupportedOperationError:
@@ -446,6 +454,7 @@ class PromptContainer(QWidget):
         Args:
             sel: Use the primary selection instead of the clipboard.
         """
+        assert self._prompt is not None
         question = self._prompt.question
         if question.url is None:
             message.error('No URL found.')
@@ -514,6 +523,7 @@ class _BasePrompt(QWidget):
         if question.text is not None:
             # Not doing any HTML escaping here as the text can be formatted
             text_label = QLabel(question.text)
+            text_label.setWordWrap(True)
             text_label.setTextInteractionFlags(Qt.TextSelectableByMouse)
             self._vbox.addWidget(text_label)
 
@@ -526,9 +536,11 @@ class _BasePrompt(QWidget):
             self.KEY_MODE.name)
         labels = []
 
+        has_bindings = False
         for cmd, text in self._allowed_commands():
             bindings = all_bindings.get(cmd, [])
             if bindings:
+                has_bindings = True
                 binding = None
                 preferred = ['<enter>', '<escape>']
                 for pref in preferred:
@@ -537,8 +549,11 @@ class _BasePrompt(QWidget):
                 if binding is None:
                     binding = bindings[0]
                 key_label = QLabel('<b>{}</b>'.format(html.escape(binding)))
-                text_label = QLabel(text)
-                labels.append((key_label, text_label))
+            else:
+                key_label = QLabel(f'<b>unbound</b> (<tt>{html.escape(cmd)}</tt>)')
+
+            text_label = QLabel(text)
+            labels.append((key_label, text_label))
 
         for i, (key_label, text_label) in enumerate(labels):
             self._key_grid.addWidget(key_label, i, 0)
@@ -549,6 +564,14 @@ class _BasePrompt(QWidget):
 
         self._vbox.addLayout(self._key_grid)
 
+        if not has_bindings:
+            label = QLabel(
+                "<b>Note:</b> You seem to have unbound all keys for this prompt "
+                f"(<tt>{self.KEY_MODE.name}</tt> key mode)."
+                "<br/>Run <tt>qutebrowser :CMD</tt> with a command from above to "
+                "close this prompt, then fix this in your config.")
+            self._vbox.addWidget(label)
+
     def _check_save_support(self, save):
         if save:
             raise UnsupportedOperationError("Saving answers is only possible "
@@ -583,6 +606,7 @@ class LineEditPrompt(_BasePrompt):
         self._vbox.addWidget(self._lineedit)
         if question.default:
             self._lineedit.setText(question.default)
+            self._lineedit.selectAll()
         self.setFocusProxy(self._lineedit)
         self._init_key_label()
 
@@ -593,7 +617,7 @@ class LineEditPrompt(_BasePrompt):
         return True
 
     def _allowed_commands(self):
-        return [('prompt-accept', 'Accept'), ('leave-mode', 'Abort')]
+        return [('prompt-accept', 'Accept'), ('mode-leave', 'Abort')]
 
 
 class FilenamePrompt(_BasePrompt):
@@ -620,6 +644,16 @@ class FilenamePrompt(_BasePrompt):
             self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
 
         self._to_complete = ''
+        self._root_index = QModelIndex()
+
+    def _directories_hide_show_model(self):
+        """Get rid of non-matching directories."""
+        num_rows = self._file_model.rowCount(self._root_index)
+        for row in range(num_rows):
+            index = self._file_model.index(row, 0, self._root_index)
+            filename = index.data()
+            hidden = self._to_complete not in filename and filename != '..'
+            self._file_view.setRowHidden(index.row(), index.parent(), hidden)
 
     @pyqtSlot(str)
     def _set_fileview_root(self, path, *, tabbed=False):
@@ -652,8 +686,10 @@ class FilenamePrompt(_BasePrompt):
             log.prompt.exception("Failed to get directory information")
             return
 
-        root = self._file_model.setRootPath(path)
-        self._file_view.setRootIndex(root)
+        self._root_index = self._file_model.setRootPath(path)
+        self._file_view.setRootIndex(self._root_index)
+
+        self._directories_hide_show_model()
 
     @pyqtSlot(QModelIndex)
     def _insert_path(self, index, *, clicked=True):
@@ -747,23 +783,22 @@ class FilenamePrompt(_BasePrompt):
         idx = self._do_completion(idx, which)
 
         selmodel.setCurrentIndex(
-            idx, QItemSelectionModel.ClearAndSelect | QItemSelectionModel.Rows)
+            idx,
+            QItemSelectionModel.ClearAndSelect |  # type: ignore[arg-type]
+            QItemSelectionModel.Rows)
         self._insert_path(idx, clicked=False)
 
     def _do_completion(self, idx, which):
-        filename = self._file_model.fileName(idx)
-        while not filename.startswith(self._to_complete) and idx.isValid():
+        while idx.isValid() and self._file_view.isIndexHidden(idx):
             if which == 'prev':
                 idx = self._file_view.indexAbove(idx)
             else:
                 assert which == 'next', which
                 idx = self._file_view.indexBelow(idx)
-            filename = self._file_model.fileName(idx)
-
         return idx
 
     def _allowed_commands(self):
-        return [('prompt-accept', 'Accept'), ('leave-mode', 'Abort')]
+        return [('prompt-accept', 'Accept'), ('mode-leave', 'Abort')]
 
 
 class DownloadFilenamePrompt(FilenamePrompt):
@@ -772,7 +807,8 @@ class DownloadFilenamePrompt(FilenamePrompt):
 
     def __init__(self, question, parent=None):
         super().__init__(question, parent)
-        self._file_model.setFilter(QDir.AllDirs | QDir.Drives | QDir.NoDot)
+        self._file_model.setFilter(
+            QDir.AllDirs | QDir.Drives | QDir.NoDotAndDotDot)  # type: ignore[arg-type]
 
     def accept(self, value=None, save=False):
         done = super().accept(value, save)
@@ -783,7 +819,7 @@ class DownloadFilenamePrompt(FilenamePrompt):
 
     def download_open(self, cmdline, pdfjs):
         if pdfjs:
-            target = downloads.PDFJSDownloadTarget()
+            target: 'downloads._DownloadTarget' = downloads.PDFJSDownloadTarget()
         else:
             target = downloads.OpenFileDownloadTarget(cmdline)
 
@@ -794,7 +830,8 @@ class DownloadFilenamePrompt(FilenamePrompt):
     def _allowed_commands(self):
         cmds = [
             ('prompt-accept', 'Accept'),
-            ('leave-mode', 'Abort'),
+            ('mode-leave', 'Abort'),
+            ('rl-filename-rubout', "Go to parent directory"),
             ('prompt-open-download', "Open download"),
             ('prompt-open-download --pdfjs', "Open download via PDF.js"),
             ('prompt-yank', "Yank URL"),
@@ -858,7 +895,7 @@ class AuthenticationPrompt(_BasePrompt):
 
     def _allowed_commands(self):
         return [('prompt-accept', "Accept"),
-                ('leave-mode', "Abort")]
+                ('mode-leave', "Abort")]
 
 
 class YesNoPrompt(_BasePrompt):
@@ -920,7 +957,7 @@ class YesNoPrompt(_BasePrompt):
             default = 'yes' if self.question.default else 'no'
             cmds.append(('prompt-accept', "Use default ({})".format(default)))
 
-        cmds.append(('leave-mode', "Abort"))
+        cmds.append(('mode-leave', "Abort"))
         cmds.append(('prompt-yank', "Yank URL"))
         return cmds
 
@@ -949,6 +986,5 @@ def init():
     """Initialize global prompt objects."""
     global prompt_queue
     prompt_queue = PromptQueue()
-    objreg.register('prompt-queue', prompt_queue)  # for commands
-    message.global_bridge.ask_question.connect(
+    message.global_bridge.ask_question.connect(  # type: ignore[call-arg]
         prompt_queue.ask_question, Qt.DirectConnection)
